"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var REQUIRED_OVERLAP = 3 / 4;
function canBeCombined(c1, c2) {
    var p1 = c1.properties;
    var p2 = c2.properties;
    if (p1.size < p2.size * REQUIRED_OVERLAP || p2.size < p1.size * REQUIRED_OVERLAP) {
        return false;
    }
    var larger;
    var smaller;
    if (p1.size > p2.size) {
        larger = p1;
        smaller = p2;
    }
    else {
        larger = p2;
        smaller = p1;
    }
    var minOverlap = Math.ceil(larger.size * REQUIRED_OVERLAP);
    var maxFaults = smaller.size - minOverlap;
    Support_1.assert(maxFaults >= 0, "Max faults negative");
    var commonProperties = [];
    var faults = 0;
    smaller.forEach(function (_, name) {
        if (larger.has(name)) {
            commonProperties.push(name);
        }
        else {
            faults += 1;
            if (faults > maxFaults)
                return false;
        }
    });
    if (faults > maxFaults)
        return false;
    for (var _i = 0, commonProperties_1 = commonProperties; _i < commonProperties_1.length; _i++) {
        var name = commonProperties_1[_i];
        var ts = smaller.get(name);
        var tl = larger.get(name);
        if (ts === undefined || tl === undefined) {
            return Support_1.panic("Both of these should have this property");
        }
        ts = Type_1.removeNull(ts);
        tl = Type_1.removeNull(tl);
        // Removing null can make unions not referentially equal.
        // We allow null properties to unify with any other.
        // FIXME: Allow some type combinations to unify, like different enums,
        // enums with strings, integers with doubles, maps with objects of
        // the correct type.
        if (ts.kind !== "null" && tl.kind !== "null" && !ts.equals(tl)) {
            return false;
        }
    }
    return true;
}
function isPartOfClique(c, clique) {
    for (var _i = 0, clique_1 = clique; _i < clique_1.length; _i++) {
        var cc = clique_1[_i];
        if (!canBeCombined(c, cc)) {
            return false;
        }
    }
    return true;
}
function makeCliqueClass(clique) {
    var result = new Type_1.ClassType(immutable_1.OrderedSet(), true);
    var _loop_1 = function (c) {
        c.names.forEach(function (n) { return result.addName(n, c.areNamesInferred); });
    };
    for (var _i = 0, clique_2 = clique; _i < clique_2.length; _i++) {
        var c = clique_2[_i];
        _loop_1(c);
    }
    return result;
}
function combineClasses(graph) {
    var unprocessedClasses = Type_1.allNamedTypesSeparated(graph).classes.toArray();
    var cliques = [];
    while (unprocessedClasses.length > 0) {
        var classesLeft = [];
        var clique = [unprocessedClasses[0]];
        for (var i = 1; i < unprocessedClasses.length; i++) {
            var c = unprocessedClasses[i];
            if (isPartOfClique(c, clique)) {
                clique.push(c);
            }
            else {
                classesLeft.push(c);
            }
        }
        if (clique.length > 1) {
            cliques.push(clique);
        }
        unprocessedClasses = classesLeft;
    }
    var combinedCliques = [];
    var replacements = immutable_1.Map();
    for (var _i = 0, cliques_1 = cliques; _i < cliques_1.length; _i++) {
        var clique = cliques_1[_i];
        var combined = makeCliqueClass(clique);
        combinedCliques.push(combined);
        for (var _a = 0, clique_3 = clique; _a < clique_3.length; _a++) {
            var c = clique_3[_a];
            replacements = replacements.set(c, combined);
        }
    }
    var replaceClasses = function (t) {
        if (t instanceof Type_1.ClassType) {
            if (combinedCliques.indexOf(t) >= 0)
                return t;
            var c = replacements.get(t);
            if (c)
                return c;
        }
        return t.map(replaceClasses);
    };
    var setCliqueProperties = function (combined, clique) {
        var properties = immutable_1.OrderedMap();
        for (var _i = 0, clique_4 = clique; _i < clique_4.length; _i++) {
            var c = clique_4[_i];
            c.properties.forEach(function (t, name) {
                var p = properties.get(name);
                if (p) {
                    p[1] += 1;
                    // If one of the clique class's properties is nullable,
                    // the combined property must be nullable, too, so we
                    // just set it to this one.  Of course it can't be one of
                    // the properties that is just null.
                    if (t.kind !== "null") {
                        p[0] = t;
                    }
                    if (t.isNullable) {
                        p[2] = true;
                    }
                }
                else {
                    properties = properties.set(name, [t, 1, t.isNullable]);
                }
            });
        }
        combined.setProperties(properties.map(function (_a, name) {
            var t = _a[0], count = _a[1], haveNullable = _a[2];
            t = replaceClasses(t);
            if (haveNullable || count < clique.length) {
                if (t.isNamedType()) {
                    t = Type_1.makeNullable(t, t.names, t.areNamesInferred);
                }
                else {
                    t = Type_1.makeNullable(t, name, true);
                }
            }
            return t;
        }));
    };
    for (var i = 0; i < cliques.length; i++) {
        setCliqueProperties(combinedCliques[i], cliques[i]);
    }
    return graph.map(replaceClasses);
}
exports.combineClasses = combineClasses;
