import { OrderedSet, OrderedMap, Map, Collection } from "immutable";
export declare type TopLevels = Map<string, Type>;
export declare type PrimitiveTypeKind = "any" | "null" | "bool" | "integer" | "double" | "string";
export declare type NamedTypeKind = "class" | "enum" | "union";
export declare type TypeKind = PrimitiveTypeKind | NamedTypeKind | "array" | "map";
export declare abstract class Type {
    readonly kind: TypeKind;
    constructor(kind: TypeKind);
    isNamedType(): this is NamedType;
    readonly abstract children: OrderedSet<Type>;
    directlyReachableTypes<T>(setForType: (t: Type) => OrderedSet<T> | null): OrderedSet<T>;
    readonly abstract isNullable: boolean;
    abstract map(f: (t: Type) => Type): Type;
    equals(other: any): boolean;
    abstract expandForEquality(other: Type): [Type, Type][] | boolean;
    abstract flatHashCode(): number;
    hashCode(): number;
}
export declare class PrimitiveType extends Type {
    readonly kind: PrimitiveTypeKind;
    constructor(kind: PrimitiveTypeKind);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(f: (t: Type) => Type): this;
    expandForEquality(other: any): boolean;
    flatHashCode(): number;
}
export declare class ArrayType extends Type {
    readonly items: Type;
    readonly kind: "array";
    constructor(items: Type);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(f: (t: Type) => Type): ArrayType;
    expandForEquality(other: Type): [Type, Type][] | boolean;
    flatHashCode(): number;
}
export declare class MapType extends Type {
    readonly values: Type;
    readonly kind: "map";
    constructor(values: Type);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(f: (t: Type) => Type): MapType;
    expandForEquality(other: Type): [Type, Type][] | boolean;
    flatHashCode(): number;
}
export declare type NameOrNames = string | OrderedSet<string>;
export declare abstract class NamedType extends Type {
    private _names;
    private _areNamesInferred;
    constructor(kind: NamedTypeKind, nameOrNames: NameOrNames, areNamesInferred: boolean);
    isNamedType(): this is NamedType;
    readonly names: OrderedSet<string>;
    readonly areNamesInferred: boolean;
    addName(name: string, isInferred: boolean): void;
    setGivenName(name: string): void;
    readonly combinedName: string;
}
export declare class ClassType extends NamedType {
    private _properties;
    kind: "class";
    constructor(names: NameOrNames, areNamesInferred: boolean, _properties?: Map<string, Type> | undefined);
    setProperties(properties: Map<string, Type>): void;
    readonly properties: Map<string, Type>;
    readonly sortedProperties: OrderedMap<string, Type>;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(f: (t: Type) => Type): ClassType;
    expandForEquality(other: Type): [Type, Type][] | boolean;
    flatHashCode(): number;
    hashCode(): number;
}
export declare class EnumType extends NamedType {
    readonly cases: OrderedSet<string>;
    kind: "enum";
    constructor(names: NameOrNames, areNamesInferred: boolean, cases: OrderedSet<string>);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(f: (t: Type) => Type): this;
    expandForEquality(other: any): boolean;
    flatHashCode(): number;
}
export declare class UnionType extends NamedType {
    readonly members: OrderedSet<Type>;
    kind: "union";
    constructor(names: NameOrNames, areNamesInferred: boolean, members: OrderedSet<Type>);
    findMember: (kind: TypeKind) => Type | undefined;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(f: (t: Type) => Type): UnionType;
    readonly sortedMembers: OrderedSet<Type>;
    equals(other: any): boolean;
    expandForEquality(other: Type): [Type, Type][] | boolean;
    flatHashCode(): number;
    hashCode(): number;
}
export declare function removeNullFromUnion(t: UnionType): [PrimitiveType | null, OrderedSet<Type>];
export declare function nullableFromUnion(t: UnionType): Type | null;
export declare function makeNullable(t: Type, typeNames: NameOrNames, areNamesInferred: boolean): Type;
export declare function removeNull(t: Type): Type;
export declare function filterTypes<T extends Type>(predicate: (t: Type) => t is T, graph: TopLevels, childrenOfType?: (t: Type) => Collection<any, Type>): OrderedSet<T>;
export declare function allNamedTypes(graph: TopLevels, childrenOfType?: (t: Type) => Collection<any, Type>): OrderedSet<NamedType>;
export declare type SeparatedNamedTypes = {
    classes: OrderedSet<ClassType>;
    enums: OrderedSet<EnumType>;
    unions: OrderedSet<UnionType>;
};
export declare function separateNamedTypes(types: Collection<any, NamedType>): SeparatedNamedTypes;
export declare function allNamedTypesSeparated(graph: TopLevels, childrenOfType?: (t: Type) => Collection<any, Type>): SeparatedNamedTypes;
export declare function matchType<U>(t: Type, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: PrimitiveType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, enumType: (enumType: EnumType) => U, unionType: (unionType: UnionType) => U): U;
