"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeBuilder = /** @class */ (function () {
    function TypeBuilder() {
        var _this = this;
        // FIXME: make mutable?
        this._primitiveTypes = immutable_1.Map();
        this._mapTypes = immutable_1.Map();
        this._arrayTypes = immutable_1.Map();
        this._enumTypes = immutable_1.Map();
        this._classTypes = immutable_1.Map();
        this._unionTypes = immutable_1.Map();
        this.getPrimitiveType = function (kind) {
            var t = _this._primitiveTypes.get(kind);
            if (t === undefined) {
                t = new Type_1.PrimitiveType(kind);
                _this._primitiveTypes = _this._primitiveTypes.set(kind, t);
            }
            return t;
        };
        this.getEnumType = function (name, isInferred, cases) {
            var t = _this._enumTypes.get(cases);
            if (t === undefined) {
                t = new Type_1.EnumType(name, isInferred, cases);
                _this._enumTypes = _this._enumTypes.set(cases, t);
            }
            else {
                t.addName(name, isInferred);
            }
            return t;
        };
        this.getMapType = function (values) {
            var t = _this._mapTypes.get(values);
            if (t === undefined) {
                t = new Type_1.MapType(values);
                _this._mapTypes = _this._mapTypes.set(values, t);
            }
            return t;
        };
        this.getArrayType = function (items) {
            var t = _this._arrayTypes.get(items);
            if (t === undefined) {
                t = new Type_1.ArrayType(items);
                _this._arrayTypes = _this._arrayTypes.set(items, t);
            }
            return t;
        };
        this.getClassType = function (name, isInferred, properties) {
            var t = _this._classTypes.get(properties);
            if (t === undefined) {
                t = new Type_1.ClassType(name, isInferred, properties);
                _this._classTypes = _this._classTypes.set(properties, t);
            }
            else {
                t.addName(name, isInferred);
            }
            return t;
        };
        this.getUniqueClassType = function (name, isInferred, properties) {
            return new Type_1.ClassType(name, isInferred, properties);
        };
        this.getUnionType = function (name, isInferred, members) {
            var t = _this._unionTypes.get(members);
            if (t === undefined) {
                t = new Type_1.UnionType(name, isInferred, members);
                _this._unionTypes = _this._unionTypes.set(members, t);
            }
            else {
                t.addName(name, isInferred);
            }
            return t;
        };
        this.getUniqueUnionType = function (name, isInferred, members) {
            return new Type_1.UnionType(name, isInferred, members);
        };
    }
    return TypeBuilder;
}());
exports.TypeBuilder = TypeBuilder;
var UnionBuilder = /** @class */ (function () {
    function UnionBuilder(typeBuilder, typeName, isInferred) {
        var _this = this;
        this.typeBuilder = typeBuilder;
        this.typeName = typeName;
        this.isInferred = isInferred;
        this._haveAny = false;
        this._haveNull = false;
        this._haveBool = false;
        this._haveInteger = false;
        this._haveDouble = false;
        this._haveString = false;
        this._arrays = [];
        this._maps = [];
        this._classes = [];
        this._enumCaseMap = {};
        this._enumCases = [];
        this.addAny = function () {
            _this._haveAny = true;
        };
        this.addNull = function () {
            _this._haveNull = true;
        };
        this.addBool = function () {
            _this._haveBool = true;
        };
        this.addInteger = function () {
            _this._haveInteger = true;
        };
        this.addDouble = function () {
            _this._haveDouble = true;
        };
        this.addString = function () {
            if (!_this._haveString) {
                _this._haveString = true;
                _this._enumCaseMap = {};
                _this._enumCases = [];
            }
        };
        this.addArray = function (t) {
            _this._arrays.push(t);
        };
        this.addClass = function (t) {
            _this._classes.push(t);
        };
        this.addMap = function (t) {
            _this._maps.push(t);
        };
        this.addEnumCase = function (s) {
            if (_this._haveString) {
                return;
            }
            if (!Object.prototype.hasOwnProperty.call(_this._enumCaseMap, s)) {
                _this._enumCaseMap[s] = 0;
                _this._enumCases.push(s);
            }
            _this._enumCaseMap[s] += 1;
        };
        this.buildUnion = function (unique) {
            var types = [];
            if (_this._haveAny) {
                return _this.typeBuilder.getPrimitiveType("any");
            }
            if (_this._haveNull) {
                types.push(_this.typeBuilder.getPrimitiveType("null"));
            }
            if (_this._haveBool) {
                types.push(_this.typeBuilder.getPrimitiveType("bool"));
            }
            if (_this._haveDouble) {
                types.push(_this.typeBuilder.getPrimitiveType("double"));
            }
            else if (_this._haveInteger) {
                types.push(_this.typeBuilder.getPrimitiveType("integer"));
            }
            if (_this._haveString) {
                types.push(_this.typeBuilder.getPrimitiveType("string"));
            }
            else if (_this._enumCases.length > 0) {
                var maybeEnum = _this.makeEnum(_this._enumCases);
                if (maybeEnum !== null) {
                    types.push(maybeEnum);
                }
                else {
                    types.push(_this.typeBuilder.getPrimitiveType("string"));
                }
            }
            if (_this._classes.length > 0 || _this._maps.length > 0) {
                types.push(_this.makeClass(_this._classes, _this._maps));
            }
            if (_this._arrays.length > 0) {
                types.push(_this.makeArray(_this._arrays));
            }
            if (types.length === 0) {
                return _this.typeBuilder.getPrimitiveType("any");
            }
            if (types.length === 1) {
                return types[0];
            }
            var typesSet = immutable_1.OrderedSet(types);
            if (unique) {
                return _this.typeBuilder.getUniqueUnionType(_this.typeName, _this.isInferred, typesSet);
            }
            else {
                return _this.typeBuilder.getUnionType(_this.typeName, _this.isInferred, typesSet);
            }
        };
    }
    Object.defineProperty(UnionBuilder.prototype, "haveString", {
        get: function () {
            return this._haveString;
        },
        enumerable: true,
        configurable: true
    });
    return UnionBuilder;
}());
exports.UnionBuilder = UnionBuilder;
